//
// includes
// --------
#define _DEFAULT_SOURCE
#define _BSD_SOURCE
#define _GNU_SOURCE

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

// Apple does not seem to have basename available
// as a part of string.h, so import libgen if needed
#if __APPLE__
  #include <libgen.h>
#endif

#include "config.h"
//
// definitions
// -----------
#define HERMES_VERSION "0.5"
#define CTRL_KEY(k) ((k) & 0x1f)

// 
// definitions for config file
// ---------------------------

enum editorMode {CommandMode, InputMode, VisualMode};

enum editorHighlight {
  HL_NORMAL = 0,
  HL_STRING,
  HL_COMMENT,
  HL_MLCOMMENT,
  HL_KEYWORD1,
  HL_KEYWORD2,
  HL_NUMBER,
  HL_VISUAL
};

enum editorKey {
  BACKSPACE = 127,
  ARROW_LEFT = 1000,
  ARROW_RIGHT,
  ARROW_UP,
  ARROW_DOWN,
  PAGE_UP,
  PAGE_DOWN,
  DEL_KEY,
  HOME_KEY,
  END_KEY
};


//
// data
// ----

#include "syntaxhl.h"

typedef struct erow {
  int idx;
  int size;
  int rendersize;
  char *chars;
  char *render;
  unsigned char *hl;
  int hl_open_comment;
} erow;

struct editorConfig {
  enum editorMode mode;
  int rowoff;
  int coloff;
  int cx, cy, rx;
  int screenRows;
  int screenCols;
  int numRows;
  int unsaved;
  int pastelen;
  int vx, vy;
  erow *row;
  char *paste;
  char *filepath;
  char *filename;
  char statusmsg[80];
  time_t statusmsg_time;
  struct termios orig_termios;
  struct editorSyntax *syntax;
};


struct editorConfig E;



// 
// prototypes
// ----------
void editorStatusMessage(const char *fmt, ...);
void editorRefreshScreen();
char *editorPrompt(char *prompt);

//
// terminal
// --------
void die(const char *s) {
  write(STDOUT_FILENO, "\033[m", 3);
  write(STDOUT_FILENO, "\033[2J", 4);
  write(STDOUT_FILENO, "\033[H", 3);
  perror(s);
  write(STDOUT_FILENO, "\r", 1);
  exit(1);
}

void quit() {
  write(STDOUT_FILENO, "\033[m", 3);
  write(STDOUT_FILENO, "\033[2J", 4);
  write(STDOUT_FILENO, "\033[H", 3);
  exit(0);
}

void disableRawMode() {
  if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &E.orig_termios) == -1) 
    die("tcsetattr");
}

void enableRawMode() {
  if (tcgetattr(STDIN_FILENO, &E.orig_termios) == -1) die("tcgetattr");
  atexit(disableRawMode);

  struct termios raw = E.orig_termios;
  raw.c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
  raw.c_oflag &= ~(OPOST);
  raw.c_cflag |= ~(CS8);
  raw.c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
  raw.c_cc[VMIN] = 0;
  raw.c_cc[VTIME] = 1;

  if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &raw) == -1) die("tcsetattr");
}

int editorReadKey() {
  int nread;
  char c;
  while ((nread = read(STDIN_FILENO, &c, 1)) != 1) {
    if (nread == -1 && errno != EAGAIN) die("read");
  }

  if (c == '\033') {
    char seq[3];

    if (read(STDIN_FILENO, &seq[0], 1) != 1) return '\033';
    if (read(STDIN_FILENO, &seq[1], 1) != 1) return '\033';
    if (seq[0] == '[') {
      if (seq[1] >= '0' && seq[1] <= '9') {
        if (read(STDIN_FILENO, &seq[2], 1) != 1) return '\033';
        if (seq[2] == '~') {
          switch (seq[1]) {
            case '1': return HOME_KEY;
            case '3': return DEL_KEY;
            case '4': return END_KEY;
            case '5': return PAGE_UP;
            case '6': return PAGE_DOWN;
            case '7': return HOME_KEY;
            case '8': return END_KEY;
          }
        }
      } else {
        switch (seq[1]) {
          case 'A': return ARROW_UP;
          case 'B': return ARROW_DOWN;
          case 'C': return ARROW_RIGHT;
          case 'D': return ARROW_LEFT;
          case 'H': return HOME_KEY;
          case 'F': return END_KEY;
        }
      }
    } else if (seq[0] == 'O') {
      switch (seq[1]) {
        case 'H': return HOME_KEY;
        case 'F': return END_KEY;
      }
    }
    return '\033';
  }
  return c;
}

int getCursorPosition(int *rows, int *cols) {
  char buf[32];
  unsigned int i = 0;

  if (write(STDOUT_FILENO, "\033[6n", 4) != 4) return -1;
  printf("\r\n");

  while (i < sizeof(buf) - 1) {
    if (read(STDIN_FILENO, &buf[i], 1) != 1) break;
    if (buf[i] == 'R') break;
    i++;
  }
  buf[i] = '\0';
  if (buf[0] != '\033' || buf[1] != '[') return -1;
  if (scanf(&buf[2], "%d;%d", rows, cols) != 2) return -1;
  return 0;
}

int getWindowSize(int *rows, int *cols) {
  struct winsize ws;

  if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws) == -1 || ws.ws_col == 0) {
    if (write(STDOUT_FILENO, "\033[999C\033[999B", 12) != 12) return -1;
    return getCursorPosition(rows, cols);
  }

  *cols = ws.ws_col;
  *rows = ws.ws_row;
  return 0;
}

// 
// syntax highlighting
// -------------------
int is_separator(int c) {
  return isspace(c) || c == '\0' || strchr(",.()+-/*=~%<>[];{}&|", c) != NULL;
}

void editorUpdateSyntax(erow *row) {
  row->hl = realloc(row->hl, row->rendersize);
  memset(row->hl, HL_NORMAL, row->rendersize);

  if (E.syntax == NULL) return;

  char **keywords = E.syntax->keywords;

  char *scs = E.syntax->singleline_comment_start;
  char *mcs = E.syntax->multiline_comment_start;
  char *mce = E.syntax->multiline_comment_end;

  int scs_len = scs ? strlen(scs) : 0;
  int mcs_len = mcs ? strlen(mcs) : 0;
  int mce_len = mce ? strlen(mce) : 0;

  int prev_sep = 1;
  int in_string = 0;
  int in_comment = (row->idx > 0 && E.row[row->idx - 1].hl_open_comment);

  int i = 0;
  while (i < row->rendersize) {
    char c = row->render[i];
    unsigned char prev_hl = (i > 0) ? row->hl[i - 1] : HL_NORMAL;

    if (scs_len && !in_string && !in_comment) {
      if (!strncmp(&row->render[i], scs, scs_len)) {
        memset(&row->hl[i], HL_COMMENT, row->rendersize - i);
        break;
      }
    }

    if (mcs_len && mce_len && !in_string) {
      if (in_comment) {
        row->hl[i] = HL_MLCOMMENT;
        if (!strncmp(&row->render[i], mce, mce_len)) {
          memset(&row->hl[i], HL_MLCOMMENT, mce_len);
          i += mce_len;
          in_comment = 0;
          prev_sep = 1;
          continue;
        } else {
          i++;
          continue;
        }
      } else if (!strncmp(&row->render[i], mcs, mcs_len)) {
        memset(&row->hl[i], HL_MLCOMMENT, mcs_len);
        i += mcs_len;
        in_comment = 1;
        continue;
      }
    }

    if (E.syntax->flags & HL_HIGHLIGHT_STRINGS) {
      if (in_string) {
        row->hl[i] = HL_STRING;
        if (c == '\\' && i < row->rendersize) {
          row->hl[i + 1] = HL_STRING;
          i += 2;
          continue;
        }
        if (c == in_string) in_string = 0;
        i++;
        prev_sep = 1;
        continue;
      } else {
        if (c == '"' || c == '\'') {
          in_string = c;
          row->hl[i] = HL_STRING;
          i++;
          continue;
        }
      }

    }

    if (E.syntax->flags & HL_HIGHLIGHT_NUMBERS) {
      if ((isdigit(c) && (prev_sep || prev_hl == HL_NUMBER)) || 
          (c == '.' && prev_hl == HL_NUMBER)) {

        row->hl[i] = HL_NUMBER;
        i++;
        prev_sep = 0;
        continue;
      }
    }

    if (prev_sep) {
      int j;
      for (j = 0; keywords[j]; j++) {
        int klen = strlen(keywords[j]);
        int kw2 = keywords[j][klen - 1] == '|';
        if (kw2) klen--;

        if (!strncmp(&row->render[i], keywords[j], klen) &&
            is_separator(row->render[i + klen])) {
          memset(&row->hl[i], kw2 ? HL_KEYWORD2 : HL_KEYWORD1, klen);
          i += klen;
          break;
        }
        if (keywords[j] != NULL) {
          prev_sep = 0;
          continue;
        }
      }
    }

    prev_sep = is_separator(c);
    i++;
  }

  int changed = (row->hl_open_comment != in_comment);
  row->hl_open_comment = in_comment;
  if (changed && row->idx + 1 < E.numRows)
    editorUpdateSyntax(&E.row[row->idx + 1]);
}

int editorSyntaxToColor(int hl) {
  switch (hl) {
    case HL_NUMBER:    return HL_NUMBER_COLOR;
    case HL_STRING:    return HL_STRING_COLOR;
    case HL_MLCOMMENT:
    case HL_COMMENT:   return HL_COMMENT_COLOR;
    case HL_KEYWORD1:  return HL_KEYWORD1_COLOR;
    case HL_KEYWORD2:  return HL_KEYWORD2_COLOR;
    case HL_VISUAL:    return VISUAL_BG_COLOR;
    default:           return HL_DEFAULT_COLOR;
  }

}

void editorSelectSyntaxHighlight() {
  E.syntax = NULL;
  if (E.filepath == NULL) return;

  char *ext = strrchr(E.filepath, '.');

  for (unsigned int j = 0; j < HLDB_ENTRIES; j++) {
    struct editorSyntax *s = &HLDB[j];
    unsigned int i = 0;
    while (s->filematch[i]) {
      int is_ext = (s->filematch[i][0] == '.');
      if ((is_ext && ext && !strcmp(ext, s->filematch[i])) ||
          (!is_ext && strstr(E.filepath, s->filematch[i]))) {
        E.syntax = s;

        int filerow;
        for (filerow = 0; filerow < E.numRows; filerow++) {
          editorUpdateSyntax(&E.row[filerow]);
        }
        return;
      }
      i++;
    }
  }
}

//
// row ops
// -------
int editorRowCxToRx(erow *row, int cx) {
  int rx = 0;
  int j;
  for (j = 0; j < cx; j++) {
    if (row->chars[j] == '\t')
      rx += (tabwidth - 1) - (rx % tabwidth);
    rx++;
  }
  return rx;
}

void editorUpdateRow(erow *row) {
  int tabs = 0;
  int j;
  for (j = 0;  j < row->size; j++)
    if (row->chars[j] == '\t') tabs++;

  free(row->render);
  row->render = malloc(row->size + tabs * (tabwidth - 1) + 1);

  int idx = 0;
  for (j = 0; j < row->size; j++) {
    if (row->chars[j] == '\t') {
      row->render[idx++] = ' ';
      while (idx % tabwidth != 0) row->render[idx++] = ' ';
    } else {
      row->render[idx++] = row->chars[j];
    }
  }
  row->render[idx] = '\0';
  row->rendersize = idx;

  editorUpdateSyntax(row);
}

void editorInsertRow(int at, char *s, size_t len) {
  if (at < 0 || at > E.numRows) return;

  E.row = realloc(E.row, sizeof(erow) * (E.numRows + 1));
  memmove(&E.row[at + 1], &E.row[at], sizeof(erow) * (E.numRows - at));
  for (int j = at + 1; j <= E.numRows; j++) E.row[j].idx++;

  E.row[at].idx = at;

  E.row[at].size = len;
  E.row[at].chars = malloc(len + 1);
  memcpy(E.row[at].chars, s, len);
  E.row[at].chars[len] = '\0';

  E.row[at].rendersize = 0;
  E.row[at].render = NULL;
  E.row[at].hl = NULL;
  E.row[at].hl_open_comment = 0;
  editorUpdateRow(&E.row[at]);

  E.numRows++;
  E.unsaved++;
}

void editorFreeRow(erow *row) {
  free(row->render);
  free(row->chars);
  free(row->hl);
}

void editorDelRow(int at) {
  if (at < 0 || at >= E.numRows) return;
  editorFreeRow(&E.row[at]);
  memmove(&E.row[at], &E.row[at + 1], sizeof(erow) * (E.numRows - at - 1));
  for (int j = at; j < E.numRows - 1; j++) E.row[j].idx--;
  E.numRows--;
  E.unsaved++;
}

void editorRowInsertChar(erow *row, int at, int c) {
  if (at < 0 || at > row->size) at = row->size;
  row->chars = realloc(row->chars, row->size+2);
  memmove(&row->chars[at + 1], &row->chars[at], row->size - at + 1);
  row->size++;
  row->chars[at] = c;
  editorUpdateRow(row);
  E.unsaved++;
}

void editorRowAppendString(erow *row, char *s, size_t len) {
  row->chars = realloc(row->chars, row->size + len + 1);
  memcpy(&row->chars[row->size], s, len);
  row->size += len;
  row->chars[row->size] = '\0';
  editorUpdateRow(row);
  E.unsaved++;
}

void editorRowDeleteChar(erow *row, int at) {
  if (at < 0 || at >= row->size) return;
  memmove(&row->chars[at], &row->chars[at + 1], row->size - at);
  row->size--;
  editorUpdateRow(row);
  E.unsaved++;
}

// 
// editor ops
// ----------
void editorInsertChar(int c) {
  if (E.cy ==  E.numRows) {
    editorInsertRow(E.numRows, "", 0);
  }
  editorRowInsertChar(&E.row[E.cy], E.cx, c);
  E.cx++;
}

void editorDeleteChar() {
  if (E.cy == E.numRows) return;
  if (E.cy == 0 && E.cx ==0) return;
  erow *row = &E.row[E.cy];
  if (E.cx > 0) {
    editorRowDeleteChar(row, E.cx - 1);
    E.cx--;
  } else {
    E.cx = E.row[E.cy - 1].size;
    editorRowAppendString(&E.row[E.cy - 1], row->chars, row->size);
    editorDelRow(E.cy);
    E.cy--;
  }
}

void editorInsertNewline() {
  if (E.cx == 0) {
    editorInsertRow(E.cy, "", 0);
  } else {
    erow *row = &E.row[E.cy];
    editorInsertRow(E.cy + 1, &row->chars[E.cx], row->size - E.cx);
    row = &E.row[E.cy];
    row->size = E.cx;
    row->chars[row->size] = '\0';
    editorUpdateRow(row);
  }
  E.cy++;
  E.cx = 0;
}

void editorUpdatePaste(char *str, int len) {
  E.paste = realloc(E.paste, len);
  memcpy(E.paste, str, len);
  E.pastelen = len;
}

//
// file io
// -------

char *editorRowsToString(int *buflen) {
  int totlen = 0;
  int j;
  for (j = 0; j < E.numRows; j++)
    totlen += E.row[j].size + 1;
  *buflen = totlen;

  char *buf = malloc(totlen);
  char *p = buf;
  for (j = 0; j < E.numRows; j++) {
    memcpy(p, E.row[j].chars, E.row[j].size);
    p += E.row[j].size;
    *p = '\n';
    p++;
  }

  return buf;
}

void editorOpen(char *filepath) {
  free(E.filepath);
  E.filepath = strdup(filepath);
  E.filename = basename(strdup(filepath));

  editorSelectSyntaxHighlight();

  FILE *fp = fopen(filepath, "a+");
  if (!fp) die("Open file");
  fseek(fp, 0, SEEK_SET);

  char *line = NULL;
  size_t linecap = 0;
  ssize_t linelen;
  while ((linelen = getline(&line, &linecap, fp)) != -1) {
    while (linelen > 0 && (line[linelen - 1] == '\n' || line[linelen - 1] == '\r'))
      linelen--;
    editorInsertRow(E.numRows, line, linelen);

  }

  free(line);
  fclose(fp);
  E.unsaved = 0;
}

void editorSave(int saveAs) {
  if (E.filepath == NULL || saveAs) {
    char *name = editorPrompt("Save as: %s");
    if (name == NULL) {
      editorStatusMessage("Save cancelled");
      return;
    }
    free(E.filepath);
    E.filepath = strdup(name);
    free(E.filename);
    E.filename = basename(E.filepath);
    editorSelectSyntaxHighlight();
  }

  int len;
  char *buf = editorRowsToString(&len);

  int fd = open(E.filepath, O_RDWR | O_CREAT, 0644);
  if (fd != -1) {
    if (ftruncate(fd, len) != -1) {
      if (write(fd, buf, len) == len) {
        close(fd);
        free(buf);
        E.unsaved = 0;
        editorStatusMessage("%d bytes saved to disk", len);
        return;
      }
    }
    close(fd);
  }
  free(buf);
  editorStatusMessage("Can't save! i/o error: %s", strerror(errno));
}

//
// append buffer
// -------------
struct abuf {
  char *b;
  int len;
};

#define ABUF_INIT {NULL, 0}

void abAppend(struct abuf *ab, const char *s, int len) {
  char *new = realloc(ab->b, ab->len + len);

  if (new == NULL) return;
  memcpy(&new[ab->len], s, len);
  ab->b = new;
  ab->len += len;
}

void abFree(struct abuf *ab) {
  free(ab->b);
}

//
// output
// ------
void editorScroll() {
  E.rx = E.cx;
  if (E.cy < E.numRows) {
    E.rx = editorRowCxToRx(&E.row[E.cy], E.cx);
  }

  if (E.cy < E.rowoff) {
    E.rowoff = E.cy;
  }
  if (E.cy >= E.rowoff + E.screenRows) {
    E.rowoff = E.cy - E.screenRows + 1;
  }
  if (E.rx < E.coloff) {
    E.coloff = E.rx;
  }
  if (E.rx >= E.coloff + E.screenCols) {
    E.coloff = E.rx - E.screenCols + 1;
  }
}

void editorAdjustCursorX() {
  erow *row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int rowlen = row ? row->size : 0;
  if (E.cx > rowlen) {
    E.cx = rowlen;
    if (E.mode == CommandMode && rowlen > 0) E.cx--;
  }

}

void editorDrawVisualBg(struct abuf *ab, int end, int drawx, int drawy, int rowlen) {
  int cond = end ? ((drawx + E.coloff == E.cx || E.cx == rowlen) && drawy == E.cy) :
    (drawx + E.coloff == E.vx && drawy == E.vy);

  if (E.mode == VisualMode && cond) {
    int color = end ? EDITOR_BG_COLOR : VISUAL_BG_COLOR;
    if (E.cy < E.vy || (E.cx < E.vx && E.cy == E.vy)) 
      color = end ? VISUAL_BG_COLOR : EDITOR_BG_COLOR;
    char vbuf[16];
    int vclen = snprintf(vbuf, sizeof(vbuf), "\033[48;5;%dm", color);
    abAppend(ab, vbuf, vclen);
  }
}

void editorDrawRows(struct abuf *ab) {
  int y;
  for (y = 0; y < E.screenRows; y++) {
    int filerow = y + E.rowoff;
    char cbuf[16];
    int cbuflen = snprintf(cbuf, sizeof(cbuf), "\033[38;5;%dm", HL_DEFAULT_COLOR);
    abAppend(ab, cbuf, cbuflen);
    if (filerow >= E.numRows) {
      if (E.numRows == 0 && y == E.screenRows / 3) {
        char welcome[80];
        int welcomelen = snprintf(welcome, sizeof(welcome), "[ hermes ] - version %s", HERMES_VERSION);
        if (welcomelen > E.screenCols) welcomelen = E.screenCols;
        int padding = (E.screenCols - welcomelen) / 2;
        if (padding) {
          abAppend(ab, "~", 1);
          padding--;
        }
        if (E.filename == NULL) {
          while (padding--) abAppend(ab, " ", 1); 
          abAppend(ab, welcome, welcomelen);
        }
      } else {
        abAppend(ab, "~", 1);
      }
    } else {
      int len = E.row[filerow].rendersize - E.coloff;
      if (len < 0) len = 0;
      if (len > E.screenCols) len = E.screenCols;
      char *c = &E.row[filerow].render[E.coloff];
      unsigned char *hl = &E.row[filerow].hl[E.coloff];
      int current_color = -1;
      if (E.mode == VisualMode && len <= 0 && filerow == E.cy) {
        int col = EDITOR_BG_COLOR;
        if (E.cy < E.vy || (E.cx < E.vx && E.cy == E.vy)) col = VISUAL_BG_COLOR;
        char vbuf[16];
        int vclen = snprintf(vbuf, sizeof(vbuf), "\033[48;5;%dm", col);
        abAppend(ab, vbuf, vclen);
      }
      int j;
      for (j = 0; j < len; j++) {
        if (iscntrl(c[j])) {
          char sym = (c[j] <= 26) ? '@' + c[j] : '?';
          abAppend(ab, "\033[7m", 4);
          abAppend(ab, &sym, 1);
          abAppend(ab, "\033[m", 3);
          char bgcolor[16];
          int bglen = sprintf(bgcolor, "\033[48;5;%dm", EDITOR_BG_COLOR);
          abAppend(ab, bgcolor, bglen);
          if (current_color != -1) {
            char buf[16];
            int clen = snprintf(buf, sizeof(buf), "\033[%dm", current_color);
            abAppend(ab, buf, clen);
          }
        } else if (hl[j] == HL_NORMAL) {
          int color = editorSyntaxToColor(hl[j]);
          if (current_color != -1) {
            current_color = -1;
            char buf[16];
            int clen = snprintf(buf, sizeof(buf), "\033[38;5;%dm", color);
            abAppend(ab, buf, clen);
          }
          editorDrawVisualBg(ab, 0, j, filerow, len);
          abAppend(ab, &c[j], 1);
          editorDrawVisualBg(ab, 1, j, filerow, len);
        } else {
          int color = editorSyntaxToColor(hl[j]);
          if (color != current_color) {
            current_color = color;
            char buf[16];
            int clen = snprintf(buf, sizeof(buf), "\033[38;5;%dm", color);
            abAppend(ab, buf, clen);
          }
          editorDrawVisualBg(ab, 0, j, filerow, len);
          abAppend(ab, &c[j], 1);
          editorDrawVisualBg(ab, 1, j, filerow, len);
        }
      }
      abAppend(ab, "\033[39m", 5);
    }

    abAppend(ab, "\033[K", 3);
    abAppend(ab, "\r\n", 2);
  }
}

void editorDrawStatusBar(struct abuf *ab) {
  char bgcolor[16];
  int bglen;
  char fgcolor[16];
  int fglen;
  char status[80], rstatus[80];

  char *editorMode;
  switch (E.mode) {
    case CommandMode:
      editorMode = "Command";
      bglen = sprintf(bgcolor, "\033[48;5;%dm", BAR_BG_COMMAND);
      fglen = snprintf(fgcolor, sizeof(fgcolor), "\033[38;5;%dm", BAR_FG_COMMAND);
      break;
    case InputMode:
      editorMode = "Input";
      bglen = sprintf(bgcolor, "\033[48;5;%dm", BAR_BG_INPUT);
      fglen = snprintf(fgcolor, sizeof(fgcolor), "\033[38;5;%dm", BAR_FG_INPUT);
      break;
    case VisualMode:
      editorMode = "Visual";
      bglen = sprintf(bgcolor, "\033[48;5;%dm", BAR_BG_VISUAL);
      fglen = snprintf(fgcolor, sizeof(fgcolor), "\033[38;5;%dm", BAR_FG_VISUAL);
      break;
  }

  abAppend(ab, fgcolor, fglen);
  abAppend(ab, bgcolor, bglen);
  
  int len = snprintf(status, sizeof(status), "%.20s - %d lines %s",
      E.filename ? E.filename : "[Unnamed]", E.numRows, E.unsaved ? "(modified)" : "");
  int rlen = snprintf(rstatus, sizeof(rstatus), "<%s>   %d:%d/%d [%s] ",
      E.syntax ? E.syntax->filetype : "generic" ,E.cx < 0 ? 0 : E.cx, E.cy + 1, E.numRows, editorMode);
  if (len > E.screenCols) len = E.screenCols;
  abAppend(ab, status, len);
  while (len < E.screenCols) {
    if (E.screenCols - len == rlen) {
      abAppend(ab, rstatus, rlen);
      break;
    } else {
      abAppend(ab, " ", 1);
      len++;
    }
  }
  abAppend(ab, "\033[m", 4);
  bglen = sprintf(bgcolor, "\033[48;5;%dm", EDITOR_BG_COLOR);
  fglen = snprintf(fgcolor, sizeof(fgcolor), "\033[38;5;%dm", HL_DEFAULT_COLOR);
  abAppend(ab, bgcolor, bglen);
  abAppend(ab, fgcolor, fglen);
  abAppend(ab, "\r\n", 2);
}

void editorDrawMsgBar(struct abuf *ab) {
  abAppend(ab, "\033[K", 3);
  char bgcolor[16];
  int bglen = sprintf(bgcolor, "\033[48;5;%dm", EDITOR_BG_COLOR);
  char fgcolor[16];
  int fglen = snprintf(fgcolor, sizeof(fgcolor), "\033[38;5;%dm", HL_DEFAULT_COLOR);
  abAppend(ab, fgcolor, fglen);
  abAppend(ab, bgcolor, bglen);
  int msglen = strlen(E.statusmsg);
  if (msglen > E.screenCols) msglen = E.screenCols;
  if (msglen && time(NULL) - E.statusmsg_time < 5)
    abAppend(ab, E.statusmsg, msglen);
}

void editorRefreshScreen() {
  editorAdjustCursorX();
  editorScroll();

  struct abuf ab = ABUF_INIT;

  abAppend(&ab, "\033[?25l", 6);
  abAppend(&ab, "\033[H", 3);
  char bgcolor[16];
  int bglen = sprintf(bgcolor, "\033[48;5;%dm", EDITOR_BG_COLOR);
  abAppend(&ab, bgcolor, bglen);
  editorDrawRows(&ab);
  editorDrawStatusBar(&ab);
  editorDrawMsgBar(&ab);

  char buf[32];
  snprintf(buf, sizeof(buf), "\033[%d;%dH", (E.cy - E.rowoff) + 1, (E.rx - E.coloff) + 1);
  abAppend(&ab, buf, strlen(buf));

  abAppend(&ab, "\033[?25h", 6);

  write(STDOUT_FILENO, ab.b, ab.len);
  abFree(&ab);
}

void editorStatusMessage(const char *fmt, ...) {
  va_list ap;
  va_start(ap, fmt);
  vsnprintf(E.statusmsg, sizeof(E.statusmsg), fmt, ap);
  va_end(ap);
  E.statusmsg_time = time(NULL);
}

void editorPrevWord() {
  int x = E.cx;
  int y = E.cy;
  int moved = 0;
  erow *row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];

  while (row) {
    if (moved && row->chars[x] != ' ' && row->chars[x] != '\t' && (x == 0 || row->chars[x - 1] == ' ' || row->chars[x - 1] == '\t')) {
      // EXIT cond
      E.cx = x;
      E.cy = y;
      break;
    } else if (x == 0) {
      y--;
      row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];
      if (!row) return;
      x = row->size;
      moved++;
    } else {
      x--;
      moved++;
    }
  }
}

void editorNextWord() {
  int ws = 0;
  int x = E.cx;
  int y = E.cy;
  erow *row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];

  while (row) {
    if (ws > 0 && row->chars[x] != ' ' && row->chars[x] != '\t') {
      E.cx = x;
      E.cy = y;
      break;
    } else if (row->chars[x] == ' ' || row->chars[x] == '\t') {
      x++;
      ws++;
    } else if (x + 1 >= row->size) {
      y++;
      x = 0;
      ws++;
      row = (y >= E.numRows || y < 0) ? NULL : &E.row[y];
      if (!row) return;
    } else {
      x++;
    }
  }
}

void editorEndOfWord() {
  int x = E.cx;
  int y = E.cy;
  erow *row = (y >= E.numRows) ? NULL : &E.row[y];

  while (x + 1 >= row->size) {
    if (y + 1 < E.numRows) {
      x = 0;
      y++;
      row = &E.row[y];
    }
  }

  if (row->chars[x + 1] == ' ' || row->chars[x + 1] == '\t') {
    x++;
  }

  while (1) {
    if (row->chars[x] != ' ' && row->chars[x] != '\t' && (x + 1 >= row->size || (row->chars[x + 1] == ' ' || row->chars[x + 1] == '\t'))) {
      E.cx = x;
      E.cy = y;
      break;
    } else {
      x++;
    }
  }
}

void editorCopyRange(int sx, int sy, int ex, int ey) {
  if (ey < sy || (sy == ey && ex < sx)) {
    int swap;
    swap = sy; sy = ey; ey = swap;
    swap = sx; sx = ex; ex = swap;
  } 
  if (sy < 0 || ey > E.numRows) return;

  int totlen = 0;
  char *buf;
  if (ey == sy) {
    totlen = ex - sx + 1;
    buf = malloc(totlen);
    memcpy(buf, &E.row[sy].chars[sx], totlen);
  } else {
    for (int y = sy; y <= ey; y++) {
      totlen += E.row[y].size + 1;
    }

    totlen -= sx;
    totlen -= (E.row[ey].size - ex);
    totlen++;

    buf = malloc(totlen);
    char *p = buf;
    for (int row = sy; row <= ey; row++) {
      if (row == sy) {
        memcpy(p, &E.row[row].chars[sx], E.row[row].size - sx);
        p += E.row[row].size - sx;
        *p = '\r';
        p++;
      } else if (row == ey) {
        memcpy(p, E.row[row].chars, ex + 1);
        p += ex + 1;
        *p = '\r';
      } else {
        memcpy(p, E.row[row].chars, E.row[row].size);
        p += E.row[row].size;
        *p = '\r';
        p++;
      }
    }
  }
  editorUpdatePaste(buf, totlen);
  free(buf);
}

void editorDeleteSelection() {
  int sx = E.cx;
  int sy = E.cy;
  int ex = E.vx;
  int ey = E.vy;
  if (E.cy < E.vy || (E.cx < E.vx && E.cy == E.vy)) {
    sx = E.vx;
    sy = E.vy;
    ex = E.cx;
    ey = E.cy;
    E.cx = sx;
    E.cy = sy;
  }
  editorCopyRange(sx, sy, ex, ey);

  if (sx < E.row[sy].size)
    E.cx++;
  while (E.cx != ex || E.cy != ey) {
    editorDeleteChar();
  }
}

//
// input
// -----
char *editorPrompt(char *prompt) {
  size_t bufsize = 128;
  char *buf = malloc(bufsize);

  size_t buflen = 0;
  buf[0] = '\0';

  while (1) {
    editorStatusMessage(prompt, buf);
    editorRefreshScreen();

    int c = editorReadKey();
    if (c == DEL_KEY || c == CTRL_KEY('h') || c == BACKSPACE) {
      if (buflen != 0) buf[--buflen] = '\0';
    } else if (c == '\033') {
      editorStatusMessage("");
      free(buf);
      return NULL;
    } else if (c == '\r') {
      if (buflen != 0) {
        editorStatusMessage("");
        return buf;
      } else {
        editorStatusMessage("");
        free(buf);
        return NULL;
      }
    } else if (!iscntrl(c) && c < 128) {
      if (buflen == bufsize - 1) {
        bufsize *= 2;
        buf = realloc(buf, bufsize);
      }
      buf[buflen++] = c;
      buf[buflen] = '\0';
    }
  }
}

void editorMoveCursor(int key) {
  erow *row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int mod = E.mode != InputMode ? 1 : 0;
  switch (key) {
    case 'j':
    case ARROW_DOWN:
      if (E.cy < E.numRows - 1) {
        E.cy++;
      }
      if (sizeof(row) > 0 && E.cx < 0) E.cx = 0;
      break;
    case 'k':
    case ARROW_UP:
      if (E.cy != 0) {
        E.cy--;
      }
      if (sizeof(row) > 0 && E.cx < 0) E.cx = 0;
      break;
    case 'h':
    case ARROW_LEFT:
      if (E.cx > 0) {
        E.cx--;
      }
      break;
    case 'l':
    case ARROW_RIGHT:
      {
        if (row && E.cx < row->size - mod)
          E.cx++;
        break;
      }
  }

  row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int rowlen = row ? row->size : 0;
  if (E.cx > rowlen - mod)
    E.cx = rowlen - mod;
}

void editorPageCursor(int key) {
  if (key == PAGE_DOWN) {
    if (E.numRows < E.screenRows) {
      E.cy = E.numRows - 1;
      return;
    }
    E.rowoff = E.rowoff + E.screenRows;
    if (E.rowoff > E.numRows - E.screenRows)
      E.rowoff = E.numRows - E.screenRows;
    E.cy = E.rowoff + 1;
  } else if (key == PAGE_UP) {
    if (E.cy < E.screenRows) {
      E.cy = 0;
      return;
    }
    E.rowoff = E.rowoff - E.screenRows;
    if (E.rowoff < 0)
      E.rowoff = 0;
    E.cy = E.rowoff + E.screenRows - 1;
  }

  erow *row = (E.cy >= E.numRows) ? NULL : &E.row[E.cy];
  int rowlen = row ? row->size : 0;
  if (E.cx > rowlen)
    E.cx = rowlen;
}

void editorInputKp(int c) {
  switch (c) {
    case '\033':
    case CTRL_KEY('l'):
      editorMoveCursor(ARROW_LEFT);
      E.mode = CommandMode;
      break;
    case '\r':
      editorInsertNewline();
      break;
    case BACKSPACE:
    case DEL_KEY:
    case CTRL_KEY('h'):
      if (c == DEL_KEY) editorMoveCursor(ARROW_RIGHT);
      editorDeleteChar();
      break;
    default:
      editorInsertChar(c);
      break;
  }
}

void editorCommandKp(int c) {
  static int deleting = 0;
  static int counter = 1;
  static int counting = 0;
  switch (c) {
    case 'j':
    case 'k':
    case 'h':
    case 'l':
      {
        int vx = E.cx;
        int vy = E.cy;
        while (counter > 0) {
          editorMoveCursor(c);
          counter--;
        }
        if (!deleting) break;
        E.vx = vx;
        E.vy = vy;
        if (c == 'h' && E.vx > 0) E.vx--; 
        if (c == 'j') {
          E.vx = 0;
          E.cx = E.row[E.cy].size - 1;
        }
        if (c == 'k') {
          E.vx = E.row[E.vy].size - 1;
          E.cx = 0;
        }
        if (c == 'l' && E.cx > 0) E.cx--;
        editorDeleteSelection();
        if (c == 'k' || c == 'j') editorDelRow(E.cy);
      }
      break;
    case '\033':
      if (E.mode == VisualMode) E.mode = CommandMode;
      break;
    case 'r':
      {
        int c = editorReadKey();
        if (c) {
          if (c == '\033') break;
          E.mode = InputMode;
          editorMoveCursor(ARROW_RIGHT);
          editorDeleteChar();
          editorInputKp(c);
          E.mode = CommandMode;
          editorMoveCursor(ARROW_LEFT);
        }
      }
      break;
    case 'g':
      {
        int vx = E.cx;
        int vy = E.cy;
        E.cy = 0;
        E.cx = 0;
        E.rowoff = 0;
        if (!deleting) break;
        E.vx = vx;
        E.vy = vy;
        int extraRow = 0;
        E.vx--;
        if (E.vx < 0) {
          if (E.vy < 1) break;
          E.vy--;
          E.vx = E.row[E.vy].size - 1;
          extraRow = 1;
        }
        editorDeleteSelection();
        if (extraRow) editorDelRow(0);
      }
      break;
    case 'G':
      {
        int vx = E.cx;
        int vy = E.cy;
        E.cy = E.numRows - 1;
        E.cx = E.row[E.cy].size - 1;
        if (E.cx < 0) E.cx = 0;

        if (!deleting) {
          if (E.numRows > E.screenRows) E.rowoff = E.numRows - E.screenRows - 1;
          break;
        }
        E.vx = vx;
        E.vy = vy;
        editorDeleteSelection();
      }
      break;
    case '\r':
      if (counting) {
        counter--;
        if (counter > E.numRows - 1) {
          E.cy = E.numRows - 1;
        } else {
          E.cy = counter;
        }
      }
      break;
    case 'o':
      E.cx = E.row[E.cy].size;
      E.mode = InputMode;
      editorMoveCursor(ARROW_RIGHT);
      editorInsertNewline();
      break;
    case 'O':
      E.cx = 0;
      E.mode = InputMode;
      editorInsertNewline();
      editorMoveCursor(ARROW_UP);
      editorMoveCursor(ARROW_RIGHT);
      break;
    case 'x':
      while (counter > 0) {
        editorUpdatePaste(&E.row[E.cy].chars[E.cx], 1);
        E.mode = InputMode;
        editorMoveCursor(ARROW_RIGHT);
        editorDeleteChar();
        E.mode = CommandMode;
        if (E.cx == E.row[E.cy].size) editorMoveCursor(ARROW_LEFT);
        counter--;
      }
      break;
    case 'i':
    case 'a':
      E.mode = InputMode;
      if (c == 'a') editorMoveCursor('l');
      break;
    case 'd':
      if (E.numRows < 1) break;
      if (E.mode == VisualMode) {
        editorDeleteSelection();
        E.mode = CommandMode;
        break;
      } else if (!deleting) {
        deleting++;
        return;
      } else {
        int start_x = E.cx;
        int start_y = E.cy;
        int end_x, end_y;
        end_y = E.cy + counter;
        if (end_y >= E.numRows) end_y = E.numRows - 1;
        end_x = E.row[end_y].size - 1;
        editorCopyRange(0, start_y, end_x, end_y);
        while (counter > 0) {
          editorDelRow(E.cy);
          if (E.cy == E.numRows) E.cy--;
          if (E.cy < 0) E.cy = 0;
          counter--;
        }
        E.cx = start_x;
        if (E.cx >= E.row[E.cy].size) E.cx = E.row[E.cy].size - 1;
        break;
      }
    case '$':
      if (deleting) {
        int diff = E.row[E.cy].size - E.cx;
        editorUpdatePaste(&E.row[E.cy].chars[E.cx], diff);
        E.cx = E.row[E.cy].size - 1;
        if (E.cx < 0) E.cx = 0;
        while (diff > 0) {
          E.mode = InputMode;
          editorMoveCursor(ARROW_RIGHT);
          editorDeleteChar();
          E.mode = CommandMode;
          if (E.cx == E.row[E.cy].size) editorMoveCursor(ARROW_LEFT);
          diff--;
        }
      } else if (E.cy < E.numRows) {
        E.cx = E.row[E.cy].size - 1;
        if (E.cx < 0) E.cx = 0;
      }
      break;
    case 'w':
      {
        int vx = E.cx;
        int vy = E.cy;
        while (counter > 0) {
          editorNextWord();
          counter--;
        }
        if (!deleting) break;
        E.vx = vx;
        E.vy = vy;
        if (E.cx == 0) {
          if (E.cy == 0) break;
          E.cy--;
          E.cx = E.row[E.cy].size - 1;
        } else {
          E.cx--;
        }
        editorDeleteSelection();
        if (E.cx > E.row[E.cy].size - 1) E.cx = E.row[E.cy].size - 1;
      }
      break;
    case 'P':
    case 'p':
      if (E.pastelen <= 0) break;
      {
        char open = c == 'p' ? 'o' : 'O';
        int startx = E.cx;
        int multiline = 0;
        if (E.paste && E.paste[E.pastelen-1] == '\r') multiline = 1;
        if (multiline) {
          editorCommandKp(open);
        } else {
          E.mode = InputMode;
          if (open == 'o') editorMoveCursor(ARROW_RIGHT);
        }
        for (int i = 0; i < E.pastelen; i++) {
          editorInputKp(E.paste[i]);
        }
        if (multiline) {
          editorInputKp(BACKSPACE);
          E.cx = startx <= E.row[E.cy].size ? startx : E.row[E.cy].size;
        }
        E.mode = CommandMode;
        if (!multiline && c == 'p') editorMoveCursor(ARROW_LEFT);
        break;
      }
    case 'b':
      {
        int start_x = E.cx;
        int start_y = E.cy;
        if (start_x > 0) {
          start_x--;
        } else if (start_y > 0) {
          start_y--;
          start_x = E.row[start_y].size - 1;
        }
        while (counter > 0) {
          editorPrevWord();
          counter--;
        }
        if (!deleting) break;

        E.vx = start_x;
        E.vy = start_y;
        editorDeleteSelection();
      }
      break;
    case 'e':
      {
        int vx = E.cx;
        int vy = E.cy;
        while (counter > 0) {
          editorEndOfWord();
          counter--;
        }
        if (!deleting) break;
        E.vx = vx;
        E.vy = vy;
        editorDeleteSelection();
      }
      break;
    case '^':
      {
        int vx = E.cx;
        E.cx = 0;
        if (deleting) {
          E.vx = vx;
          E.vy = E.cy;
          if (E.vx > 0) E.vx--;
          editorDeleteSelection();
        }
      }
      break;
    case 'v':
    case 'V':
      if (E.mode == CommandMode) {
        E.vx = E.cx;
        E.vy = E.cy;
        E.mode = VisualMode;
      }
      break;
    case 'y':
      if (E.mode == VisualMode) {
        editorCopyRange(E.cx, E.cy, E.vx, E.vy);
        E.mode = CommandMode;
        E.cy = E.vy;
        E.cx = E.vx;
      }
      break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
      {
        int num = c - 48;
        if (!counting && counter <= 1) {
          if (num == 0) break;
          counter = num;
        } else if (counting && num == 0) {
          counter *= 10; 
        } else {
          counter = num + (counter * 10); 
        }
        counting++;
      }
      return;
  }
  deleting = 0;
  counter = 1;
  counting = 0;
}

void editorProcessKeypress() {
  static int quit_times = UNSAVED_QUIT_COUNT;
  int c = editorReadKey();
  switch (c) {
    case CTRL_KEY('q'):
      if (E.unsaved && quit_times > 0) {
        editorStatusMessage("Warning: unsaved changed. Press Ctrl-q %d more times to quit anyway", quit_times);
        quit_times--;
        return;
      }
      quit();
      break;
    case ARROW_DOWN:
    case ARROW_UP:
    case ARROW_LEFT:
    case ARROW_RIGHT:
      editorMoveCursor(c);
      return;
    case PAGE_UP:
    case PAGE_DOWN:
      editorPageCursor(c);
      return;
    case CTRL_KEY('u'):
      editorPageCursor(PAGE_UP);
      return;
    case CTRL_KEY('d'):
      editorPageCursor(PAGE_DOWN);
      return;
    case CTRL_KEY('s'):
      editorSave(0);
      return;
    case CTRL_KEY('a'):
      editorSave(1);
      return;
    case HOME_KEY:
      E.cx = 0;
      break;
    case END_KEY:
      if (E.cy < E.numRows)
        E.cx = E.row[E.cy].size;
      break;
  }

  quit_times = UNSAVED_QUIT_COUNT;

  if (E.mode == CommandMode || E.mode == VisualMode) {
    editorCommandKp(c);
  } else if (E.mode == InputMode) {
    editorInputKp(c);
  }
}

void updateWindowSize(void) {
  if (getWindowSize(&E.screenRows, &E.screenCols) == -1)
    die("Screen resize");
  E.screenRows -= 2;
}

void handleSigWinCh(int unused __attribute__((unused))) {
  updateWindowSize();
  if (E.cy > E.screenRows) E.cy = E.screenRows - 1;
  if (E.cx > E.screenCols) E.cx = E.screenCols - 1;
  // TODO does rowoff need to be set here?
  editorRefreshScreen();
}

//
// init
// ----
void initEditor() {
  E.mode           = CommandMode;
  E.rowoff         = 0;
  E.coloff         = 0;
  E.numRows        = 0;
  E.unsaved        = 0;
  E.statusmsg_time = 0;
  E.cx             = 0;
  E.cy             = 0;
  E.rx             = 0;
  E.vx             = 0;
  E.vy             = 0;
  E.paste          = NULL;
  E.row            = NULL;
  E.filepath       = NULL;
  E.filename       = NULL;
  E.syntax         = NULL;
  E.statusmsg[0]  = '\0';

  updateWindowSize();
  signal(SIGWINCH, handleSigWinCh);
}

int main(int argc, char *argv[]) {
  enableRawMode();
  initEditor();
  if (argc >= 2) {
    editorOpen(argv[1]);
  }

  editorStatusMessage("Ctrl-s to save - Ctrl-q to quit");

  while (1) {
    editorRefreshScreen();
    editorProcessKeypress();
  }
  return 0;
}
